/* ----------------------------------------------------------------
 *   FILE
 *	fe-threads.c
 *	
 *   DESCRIPTION
 *	Most of the code to make libpq thread-aware and able to handle
 *	multiple connections at a time;  the rest consists of ifdef'ed
 *	snippets sprinkled through a couple of the files that actually
 *	invoke thread-aware versions of code in this file, or that just
 *	do the right thing in-place if it's only a single line or two.
 *
 *   SUPPORT ROUTINES
 *	None.
 *
 *   INTERFACE ROUTINES
 *	_libpq_init_threads	initialize module
 *				should not actually be called from
 *				outside this file, since everything
 *				internally makes sure we are initialized
 *				via the pthread_once() mechanism.
 *	_libpq_setup_thread	initialize per-thread state for this
 *				module; again, hopefully not needed
 *				from anywhere outside of here.
 *	_libpq_update_use_time	update a connection's idea of when
 *				it was last used.
 *	_libpq_update_current_use_time
 *				update current thread's current
 *				connection's LRU time
 *	_libpq_EstablishComm	threaded version of fe-pqexec.c:EstablishComm
 *
 *   NOTES
 *	I don't directly rely on the POSIX threads interface in most of
 *	this, but rather use macros that hide it instead.  This is to
 *	make it possible to compile on top of various, brain-dead
 *	implementations with incompatible APIs, i.e. MIT vs. DEC OSF/1.
 *
 *   AUTHOR
 *	Sean Levy <snl@paneer.ndim.edrc.cmu.edu>
 *
 *   IDENTIFICATION
 *	$Id: fe-threads.c,v 1.3 1995/06/23 03:48:10 snl Exp $
 * ----------------------------------------------------------------
 */

#ifdef USE_THREADS

#define __fe_threads_c__

#include "tmp/pqthreads.h"

#include "tmp/c.h"
#include "tmp/simplelists.h"
#include "tmp/libpq.h"
#include "tmp/libpq-fe.h"
#include "tmp/libpq-fs.h"
#include "tmp/postgres.h"
#include "fmgr.h"
#include "tmp/fastpath.h"
#include "libpq/auth.h"
#include "utils/hsearch.h"
#include <stdlib.h>

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024		/* BOGUS HARDCODED CONSTANT */
#endif

typedef int libpq_connection_t;	/* magic cookie for connection */
typedef int libpq_dbname_t;	/* magic cookie for database name */

/* libpq_connection_state_t: holds the state-per-connection needed to
   allow multiple connections to exist and be used by multiple threads
   at the same time. */

typedef struct _libpq_connection_state {
  /* The top of the struct bookkeepping stuff for dealing with
     multiple threads bashing on the same connection at the same time. */

  LOCK_T _lock;			/* lock on the rest of the struct */
  THREAD_T _owner;		/* thread that currently owns the connection */
  COND_T _waiters;		/* threads waiting for a crack at it */
  int _n_waiters;		/* count of waiting threads */
  int _inuse;			/* how many threads consider this their current conn */
  int _xact_count;		/* nesting count for begin/end pairs */
  struct sockaddr_in _addr;	/* host/port of database server */
  libpq_dbname_t _dbname;	/* hashed database name */
  struct timeval _used;		/* time for LRU reaper */
  int _dont_reap;		/* dont reap me no matter what */

  /* The following are where the real values for the old-style globals
     that libpq uses are stored. */

  char *PQhost;			/* name of database server host */
  char *PQport;			/* stringified port# */
  char *PQtty;			/* where do error messages go? */
  char *PQoption;		/* options to the backend */
  char PQdatabase[17];		/* database name */
  int PQportset;		/* OBSOLETE: communications initialized yet? */
  int PQxactid;			/* transaction ID */
  int PQAsyncNotifyWaiting;	/* for broken async notification junk */
  FILE *Pfin;			/* input stream from server */
  FILE *Pfout;			/* output stream to server */
  int PQtracep;
} libpq_connection_state_t;

/* Libpq_per_thread_state_t: state squrreled away in a thread's per-thread
   state area to maintain the thread's idea of its "current" connection
   and to factor out certain disgusting global variables that really should
   be per-thread. */

typedef struct _libpq_per_thread_state {
  int cid;			/* my current connection */
  char PQerrormsg[error_msg_length+1];
				/* last error message */
  int p_errno;			/* last PQfs* error code */
  char cwdir[MAXPATHLEN];	/* for fe-pqufs.c p_xxx functions */
  char thenewdir[MAXPATHLEN];	/* for fe-pqufs.c resolve_path() fn */
  char retbuf[portal_name_length+1];
				/* random buffer for stupid API */
  FILE *debug_port;		/* debugging output stream */
} libpq_per_thread_state_t;

/* Module-global state for threads junk. */

static ONCE_T threads_init_once = ONCE_INIT;
				/* object that lets us safely invoke */
				/* the module's initialization routine */
static THREAD_STATE_KEY_T libpq_state_key;
				/* key to use for storing per-thread state */
				/* that this module needs */

static LOCK_T connection_table_lock;
				/* lock for the connections table */
static libpq_connection_state_t *connections;
				/* all known connections */
static int n_connections = 0;	/* # slots in table used */
static int max_n_connections = 0;
				/* total slots allocated for table */

static LOCK_T hosts_lock;	/* lock for hosts table */
static HTAB *hosts;		/* key:name, val:ip address */

static LOCK_T dbnames_lock;	/* lock for dbnames table */
static HTAB *dbnames;		/* key:name, val:id */
static u_long unique_dbnames;	/* counter to generate unique dbnames */

static THREAD_T conn_reaper;	/* reaper thread */

int PQpersistentConns = 0;	/* reapable connections? [stupid libpq-style API] */

unsigned PQdeadConnSecs = 30*60;/* how many seconds before we consider a connection */
				/* to be inactive and potentially reapable */

unsigned PQconnReapSleep = 30*60;
				/* how many seconds the reaper thread sleeps */
				/* in between runs */

#define _INIT_MODULE() RUN_ONCE(threads_init_once,_libpq_internal_init_threads)

#define INIT_N_CONNS	8
#define GROW_CONN_TABLE	8
#define INVALID_CONN ((libpq_connection_t)-1)
#define FREAK(x) { printf x; fflush(stdout); abort(); }

/* forwards */

static char *dup_str ARGS((char*));
static void _libpq_exit_cleanup ARGS((void));
static void _libpq_internal_init_threads ARGS((void));
void _libpq_init_threads ARGS((void));
libpq_per_thread_state_t *_libpq_setup_thread ARGS((void));
static void _libpq_cleanup_thread ARGS((void*));
static libpq_dbname_t _libpq_get_dbname ARGS((char*));
static libpq_connection_t _libpq_find_connection ARGS((char*,u_short,char*));
static libpq_connection_t _libpq_current_connection ARGS((void));
static void _libpq_zap_connection ARGS((libpq_connection_t));
static libpq_connection_t _libpq_new_connection ARGS((void));
static void _libpq_assign_connection ARGS((libpq_connection_t));
static void _libpq_release_connection ARGS((libpq_connection_t));
static void _libpq_tear_down_connection ARGS((libpq_connection_t));
static void _libpq_reap_connections ARGS((int));
void _libpq_update_use_time ARGS((libpq_connection_t));
void _libq_update_current_use_time ARGS((void));
static void *_libpq_reaper ARGS((void *));
int _libpq_EstablishComm ARGS((void));

/* code */

/* dup_str -- duplicate a string

	the ever-duplicated string-duplication primitive */

static char *
dup_str(s)
     char *s;
{
  char *new_str;

  if (!s)
    s = "";
  new_str = (char *)malloc(strlen(s) + 1);
  if (new_str)
    strcpy(new_str, s);
  return new_str;
}

/* _libpq_exit_cleanup -- cleanup on process exit

	installed via atexit() to tear down all connections
	when the process exits. */

static void
_libpq_exit_cleanup()
{
  _libpq_reap_connections(1);
}

/* _libpq_internal_init_threads -- init this module

	set up internal data structures, locks, etc. */

static void
_libpq_internal_init_threads()
{
  int i;
  HASHCTL info;

  INIT_LOCK(connection_table_lock);
  STATE_KEY_INIT(libpq_state_key, _libpq_cleanup_thread);
  connections =
    (libpq_connection_state_t *)malloc(INIT_N_CONNS*sizeof(libpq_connection_state_t));
  if (!connections) {
    fprintf(stderr, "libpq:threads module could not allocate connection table\n");
    exit(1);
  }
  max_n_connections = INIT_N_CONNS;
  for (i = 0; i < max_n_connections; i++) {
    INIT_LOCK(connections[i]._lock);
    SET_TO_NULL_THREAD(connections[i]._owner);
    INIT_COND(connections[i]._waiters);
    connections[i]._n_waiters = 0;
    connections[i]._inuse = -1;
  }
  n_connections = 0;
  info.keysize = MAXPATHLEN;
  info.datasize = sizeof(u_long);
  info.hash = string_hash;
  hosts = hash_create(16, &info, HASH_ELEM | HASH_FUNCTION);
  dbnames = hash_create(16, &info, HASH_ELEM | HASH_FUNCTION);
  if (pq_global_init() < 0) {
    FREAK(("could not initialize libpq global state"));
  }
  INIT_LOCK(hosts_lock);
  INIT_LOCK(dbnames_lock);
  FORK_THREAD(conn_reaper, _libpq_reaper, 0);
  atexit(_libpq_exit_cleanup);
}

/* _libpq_init_threads -- pthread_once init fn

	make sure this module has been initialized; invoked
	indirectly via pthread_once(). */

void
_libpq_init_threads()
{
  _INIT_MODULE();
}

/* _libpq_setup_thread -- setup per-thread state if needed

	install libpq per-thread state in the current thread if
	it doesn't have it already and return a pointer to it. */

libpq_per_thread_state_t *
_libpq_setup_thread()
{
  libpq_per_thread_state_t *s;

  s = GET_PER_THREAD_STATE(libpq_state_key);
  if (!s) {
    s = (libpq_per_thread_state_t *)malloc(sizeof(libpq_per_thread_state_t));
    if (!s) {
      FREAK(("could not allocate libpq per-thread state"));
    }
    s->cid = INVALID_CONN;
    s->PQerrormsg[0] = '\0';
    s->p_errno = 0;
    s->cwdir[0] = '/';
    s->thenewdir[0] = '\0';
    s->retbuf[0] = '\0';
    s->debug_port = 0;
    SET_PER_THREAD_STATE(libpq_state_key,(void *)s);
  }
  return s;
}

/* _libpq_cleanup_thread -- cleanup on thread death

	tear down per-thread state when a thread dies; invoked
	indirectly via the pthread cleanup mechanism. */

static void
_libpq_cleanup_thread(ptr)
     void *ptr;
{
  libpq_per_thread_state_t *s = (libpq_per_thread_state_t *)ptr;

  if (s) {
    if (s->cid != INVALID_CONN && s->cid >= 0 && s->cid < n_connections) {
      libpq_connection_t cid = s->cid;

      /* This code is cut and pasted here because by the time we're being
	 called, pthread_getspecific() will return 0x0, which screws up
	 the way tha the rest of the code is written, i.e. calling
	 _libpq_setup_thread() to get a handle on s.  We have been passed
	 in the last valid handle to s, so we need to deal with it here. */
      if (!THREAD_IS_SELF(connections[cid]._owner)) {
	FREAK(("tried to release cid#%d but don't own it",cid));
      }
      if (!connections[cid]._xact_count) {
	int n_waiters = connections[cid]._n_waiters;

	connections[cid]._inuse--;
	if (connections[cid]._inuse < 0) {
	  FREAK(("cid#%d _inuse counter underflow",cid));
	}
	s->cid = INVALID_CONN;
	SET_TO_NULL_THREAD(connections[cid]._owner);
	REL_LOCK(connections[cid]._lock);
	if (n_waiters) {
	  SIGNAL_COND(connections[cid]._waiters);
	}
      }
    }
    free((void *)s);
  }
}

/* _libpq_get_dbname -- hash database names

	look up a db name in our hashtable, so we can avoid doing
	string comparisons. */

static libpq_dbname_t
_libpq_get_dbname(db)
     char *db;
{
  char *hashPtr, temp_dbname[MAXPATHLEN];
  libpq_dbname_t result, *resultPtr;
  Boolean found = 0;

  GET_LOCK(dbnames_lock);
  bzero(temp_dbname, sizeof(temp_dbname));
  strcpy(temp_dbname, db);
  hashPtr = (char *)hash_search(dbnames, temp_dbname, HASH_ENTER, &found);
  resultPtr = (libpq_dbname_t *)(hashPtr + MAXPATHLEN);
  if (!found)
    result = *resultPtr = ++unique_dbnames;
  else
    result = *resultPtr;
  REL_LOCK(dbnames_lock);
  return result;
}

/* _libpq_find_connection -- look for existing connection

	search the table of connections for an existing connection
	to a host,port,db triple. we cache hostname=>ip address
	and dbname=>db id mappings in hash tables.

	XXX ASSSUMES WE HAVE THE TABLE LOCK XXX */

static libpq_connection_t
_libpq_find_connection(hostname, portno, db)
     char *hostname;
     u_short portno;		/* host byte order */
     char *db;
{
  char *hashPtr;
  libpq_connection_t result;
  u_long *hostaddrPtr, hostaddr;
  Boolean found = 0;
  char temp_hostname[MAXPATHLEN]; /* YUCK! */

  (void) _libpq_setup_thread();
  result = INVALID_CONN;
  GET_LOCK(hosts_lock);
  bzero(temp_hostname, sizeof(temp_hostname));
  strcpy(temp_hostname, hostname);
  hashPtr = (char *)hash_search(hosts, temp_hostname, HASH_FIND, &found);
  if (found) {
    hostaddrPtr = (u_long *)(hashPtr + MAXPATHLEN);
    hostaddr = *hostaddrPtr;
  } else {
    struct sockaddr_in tmp;

    if (pq_getinaddr(&tmp, hostname, portno))
      hostaddr = 0;
    else {
      hostaddr = tmp.sin_addr.s_addr;
      hashPtr = (char *)hash_search(hosts, temp_hostname, HASH_ENTER, &found);
      if (found) {
	FREAK(("host %s was not found and then WAS found",hostname));
      }
      bzero(hashPtr, MAXPATHLEN);
      strcpy(hashPtr, hostname);
      hostaddrPtr = (u_long *)(hashPtr + MAXPATHLEN);
      *hostaddrPtr = hostaddr;
    }
  }
  if (hostaddr) {
    int i;
    u_short nportno = htons(portno);
    libpq_dbname_t dbname;

    dbname = _libpq_get_dbname(db);
    for (i = 0; i < n_connections; i++)
      if (connections[i]._addr.sin_addr.s_addr == hostaddr &&
	  connections[i]._addr.sin_port == nportno &&
	  connections[i]._dbname == dbname)
	break;
    if (i < n_connections)
      result = i;
  }
  REL_LOCK(hosts_lock);
  return result;
}

/* _libpq_current_connection -- get thread's current connection if any

	return the id of this thread's current connection, or
	INVALID_CONN if it is not talking to any postgres database. */

static libpq_connection_t
_libpq_current_connection()
{
  libpq_per_thread_state_t *s;
  libpq_connection_t cid;

  s = (libpq_per_thread_state_t *)GET_PER_THREAD_STATE(libpq_state_key);
  cid = s? (s->cid): INVALID_CONN;
  return cid;
}

/* _libpq_zap_connection -- zero out connection table entry

	initialize all members of connection table entry to
	default values. */

static void
_libpq_zap_connection(cid)
     libpq_connection_t cid;
{
  connections[cid]._inuse = 0;
  connections[cid]._xact_count = 0;
  bzero((char *)&connections[cid]._addr, sizeof(struct sockaddr_in));
  connections[cid]._dbname = 0;
  bzero((char*)&connections[cid]._used, sizeof(struct timeval));
  connections[cid]._dont_reap = 0;

  connections[cid].PQhost = 0;
  connections[cid].PQport = 0;
  connections[cid].PQtty = 0;
  connections[cid].PQoption = 0;
  connections[cid].PQdatabase[0] = '\0';
  connections[cid].PQportset = 0;
  connections[cid].PQxactid = 0;
  connections[cid].PQAsyncNotifyWaiting = 0;
  connections[cid].Pfin = 0;
  connections[cid].Pfout = 0;
  connections[cid].PQtracep = 0;
}

/* _libpq_new_connection -- alloc conntable entry

	allocate an unused slot in the connection table, possibly
	growing it if it is full, and return the id.

	XXX ASSUMES WE HAVE THE TABLE LOCK XXX */

static libpq_connection_t
_libpq_new_connection()
{
  libpq_connection_t cid;

  if (n_connections > max_n_connections) {
    FREAK(("can't happen: n_connections(%d) > max_n_connections(%d)",n_connections,
	   max_n_connections));
  } else if (n_connections == max_n_connections) {
    int i;

    for (cid = 0; cid < n_connections; cid++)
      if (connections[cid]._inuse < 0)
	break;
    if (cid == n_connections) {
      max_n_connections += GROW_CONN_TABLE;
      connections = (libpq_connection_state_t *)realloc((void *)connections,
							max_n_connections *
						      sizeof(libpq_connection_state_t));
      if (!connections) {
	FREAK(("could not grow libpq connections table"));
      }
      for (i = n_connections; i < max_n_connections; i++) {
	INIT_LOCK(connections[cid]._lock);
	SET_TO_NULL_THREAD(connections[cid]._owner);
	INIT_COND(connections[cid]._waiters);
	connections[i]._n_waiters = 0;
	connections[i]._inuse = -1;
      }
      n_connections++;
    }
  } else
    cid = n_connections++;
  _libpq_zap_connection(cid);
  return cid;
}

/* _libpq_begin_xact -- start a transaction

	called when libpq sees a "begin" statement go by; we keep
	track of the transaction nesting level [can they even be
	nested in the current version?] to avoid problems with
	thread context switches */

void
_libpq_begin_xact()
{
  libpq_connection_t cid;

  _libpq_setup_thread();
  cid = _libpq_current_connection();
  if (cid != INVALID_CONN) {
    connections[cid]._xact_count++;
  }
}

/* _libpq_end_xact -- end a transaction

	invoked when an "end" transaction is seen. */

void
_libpq_end_xact()
{
  libpq_connection_t cid;

  _libpq_setup_thread();
  cid = _libpq_current_connection();
  if (cid != INVALID_CONN) {
    connections[cid]._xact_count--;
  }
}

/* _libpq_assign_connection -- make a connection "ours" 

	This function is coded somewhat more defensively that it
	needs to be at the moment, since the only way for the _waiters
	thing to happen is in this module.  Nevertheless, it is a good
	practise to code defensively when dealing with concurrency, and
	I don't think the slight added performance impact is worth
	worrying about right now. */

static void
_libpq_assign_connection(cid)
     libpq_connection_t cid;
{
  int got_it;
  libpq_connection_t other_cid;
  libpq_per_thread_state_t *s;

  s = _libpq_setup_thread();
  other_cid = s->cid;
  if (other_cid != INVALID_CONN) {
    if (connections[other_cid]._xact_count) {
      FREAK(("trying to assign cid#%d while in the middle of a transaction on cid#%d",
	     cid,other_cid));
    }
    _libpq_release_connection(other_cid);
  }
  got_it = 0;
  GET_LOCK(connections[cid]._lock);
  if (!connections[cid]._xact_count) {
    SET_THREAD_TO_SELF(connections[cid]._owner);
    got_it = 1;
  }
  while (!got_it) {
    connections[cid]._n_waiters++; /* XXX potential atomicity problem */
    WAIT_COND(connections[cid]._waiters, connections[cid]._lock);
    if (!connections[cid]._xact_count) {
      SET_THREAD_TO_SELF(connections[cid]._owner);
      connections[cid]._n_waiters--;
      got_it = 1;
    }
  }
  if (got_it == 1) {
    connections[cid]._inuse++;
    s->cid = cid;
  }
}

/* _libpq_release_connection -- have a thread give up its connection

	release our hold on the given connection; if it is not
	our current connection, we panic, since that mean something
	very bad happened [like interleaved connection setups that
	don't bracket with PQfinish() properly]. */

static void
_libpq_release_connection(cid)
     libpq_connection_t cid;
{
  libpq_per_thread_state_t *s;

  s = _libpq_setup_thread();
  if (s->cid != cid) {
    FREAK(("tried to release cid#%d but it is not my current connection",cid));
  }
  if (!THREAD_IS_SELF(connections[cid]._owner)) {
    FREAK(("tried to release cid#%d but don't own it",cid));
  }
  if (!connections[cid]._xact_count) {
    int n_waiters = connections[cid]._n_waiters;

    connections[cid]._inuse--;
    if (connections[cid]._inuse < 0) {
      FREAK(("cid#%d _inuse counter underflow",cid));
    }
    s->cid = INVALID_CONN;
    SET_TO_NULL_THREAD(connections[cid]._owner);
    REL_LOCK(connections[cid]._lock);
    if (n_waiters) {
      SIGNAL_COND(connections[cid]._waiters);
    }
  }
}

/* _libpq_tear_down_connection -- actually shut down a connection

	tell the backend on the other side of a given connection
	to shut down, close the TCP socket and release the
	conntable entry for reuse. */

static void
_libpq_tear_down_connection(cid)
     libpq_connection_t cid;
{
  if (cid < 0 || cid >= n_connections || connections[cid]._inuse < 0) {
    FREAK(("tear_down_connection called on bogus cid#%d",cid));
  }
  if (connections[cid]._inuse > 0) {
    FREAK(("tear_down_connection called on in-use cid#%d",cid));
  }
  /* Tell the backend to shut down */
  (void) fputc('X', connections[cid].Pfout);
  fflush(connections[cid].Pfout);
  fclose(connections[cid].Pfout);
#ifdef MIT_PTHREADS
  funlockfile(connections[cid].Pfin);
#endif /* MIT_PTHREADS */
  fclose(connections[cid].Pfin);
  if (connections[cid].PQoption)
    free((void *)connections[cid].PQoption);
  if (connections[cid].PQtty)
    free((void *)connections[cid].PQtty);
  if (connections[cid].PQport)
    free((void *)connections[cid].PQport);
  if (connections[cid].PQhost)
    free((void *)connections[cid].PQhost);
  _libpq_zap_connection(cid);
  connections[cid]._inuse = -1;	/* zap sets it to zero */
}

/* _libpq_reap_connections -- tear down multiple connections

	if called with reap_all==1, tear down all connections.
	if called with reap_all==0, tear down all open connections
	that have not been used in PQdeadConnSecs [default=30min]. */

static void
_libpq_reap_connections(reap_all)
     int reap_all;
{
  int i;
  struct timeval now;

  gettimeofday(&now, NULL);
  GET_LOCK(connection_table_lock);
  for (i = 0; i < n_connections; i++) {
    if (reap_all ||
	(!connections[i]._inuse && !connections[i]._dont_reap &&
	 ((connections[i]._used.tv_sec + PQdeadConnSecs) < now.tv_sec)))
      _libpq_tear_down_connection(i);
  }
  REL_LOCK(connection_table_lock);
}

/* _libpq_update_use_time -- update LRU time for connection

	called from various places in libpq to update the
	time-last-used for a connection, which is used by the
	connection reaper to do its LRU thing. */

void
_libpq_update_use_time(cid)
     libpq_connection_t cid;
{
  if (cid >= 0 && cid < n_connections && connections[cid]._inuse >= 0)
    gettimeofday(&connections[cid]._used, NULL);
}

/* _libpq_update_current_use_time -- update thread's current connection

	update thread's current connection's last use time. */

void
_libpq_update_current_use_time()
{
  libpq_connection_t cid;

  cid = _libpq_current_connection();
  if (cid != INVALID_CONN)
    _libpq_update_use_time(cid);
}

/* _libpq_reaper -- reaper thread main routine

	function forked off into its own thread that wakes up every
	30 minutes and reaps unused connections. */

static void *
_libpq_reaper(junk_arg)
     void  *junk_arg;
{
  while (1) {
    sleep(PQconnReapSleep);
    _libpq_reap_connections(0);
  }
}

/* _libpq_dump_threads_state -- produce debugging output

	dump out a formatted representation of the state of
	the connection table.

	XXX NOTE: this is meant to be used from the debugger
	or in code that is being debugged, but note that it
	can itself cause core dumps depending on the state of
	the locks, what version of stdio you are using, etc. XXX */

void
_libpq_dump_threads_state(fp)
     FILE *fp;
{
  int i;

  if (!fp)
    fp = stderr;
  fprintf(fp, "%d / %d connections in table@%lx:\n",n_connections,max_n_connections,
	  connections);
  for (i = 0; i < n_connections; i++) {
    fprintf(fp, "[%d]\t", i);
    if (TRY_LOCK(connections[i]._lock))
      fprintf(fp, "locked ");
    else {
      fprintf(fp, "unlocked ");
      REL_LOCK(connections[i]._lock);
    }
    fprintf(fp, "owner:");
    fprintf(fp, THREAD_PRINTF_FMT, THREAD_PRINTF_ARG(connections[i]._owner));
    fprintf(fp, " waiters:%d", connections[i]._n_waiters);
    fprintf(fp, " inuse:%d\n", connections[i]._inuse);
    if (connections[i]._inuse>=0) {
      struct timeval now;

      fprintf(fp, "\txact:%d ", connections[i]._xact_count);
      fprintf(fp, "server:<%s:%d> ",
	    inet_ntoa(connections[i]._addr.sin_addr),
	    ntohs(connections[i]._addr.sin_port));
      fprintf(fp, "db:%d ", connections[i]._dbname);
      gettimeofday(&now,NULL);
      fprintf(fp, "used:%d ",now.tv_sec-connections[i]._used.tv_sec);
      fprintf(fp, "reap?%s\n", connections[i]._dont_reap?"no":"yes");
      fprintf(fp, "\tPQhost:%s PQport:%s PQtty:%s PQoption:%s\n",
	      connections[i].PQhost, connections[i].PQport, connections[i].PQtty,
	      connections[i].PQoption);
      fprintf(fp, "\tPQdatabase=%s PQportset=%d PQxactid=%d PQANW=%d PQtracep=%d\n",
	      connections[i].PQdatabase, connections[i].PQportset,
	      connections[i].PQxactid, connections[i].PQAsyncNotifyWaiting,
	      connections[i].PQtracep);
    }
  }
}

/* _libpq_setup_connection -- connect thread to a postgres database

	if a connection to the given host,port,db triple already
	exists, wait for it to become available and allocate it
	to the current thread, otherwise open a new connection and
	allocate it to the current thread. */

int
_libpq_setup_connection(db, host, port)
     char *db;
     char *host;
     char *port;
{
  extern void PQufs_init();
  libpq_connection_t cid, other_cid;
  u_short portno;
  extern char *PQoption, *PQtty;
  libpq_per_thread_state_t *s;

  portno = atoi(port);
  s = _libpq_setup_thread();
  cid = s->cid;
  other_cid = _libpq_find_connection(host,portno,db);
  if ((cid == INVALID_CONN) || (cid != other_cid)) {
    cid = pq_connect(db,
		     fe_getauthname(),
		     PQoption,
		     host,
		     PQtty,
		     (char *) NULL,
		     (short) portno);
    if (cid == INVALID_CONN) {
      (void) sprintf(s->PQerrormsg, 
		     "FATAL: Failed to connect to postmaster (host=%s, port=%s)\n",
		     host, port);
      pqdebug("%s", s->PQerrormsg);
      (void) strcat(s->PQerrormsg, "\tIs the postmaster running?\n");
      fputs(s->PQerrormsg, stderr);
      cid = INVALID_CONN;
    } else {
      pq_flush();
      PQufs_init();
    }
  }
  if (cid != INVALID_CONN)
    _libpq_update_use_time(cid);
  return (int)cid;
}

/* _libpq_EstablishComm -- thread-aware version of EstablishComm

	called from fe-pqexec.c:EstablishComm if compiled with
	USE_THREADS turned on; this does pretty much what 
	_libpe_setup_connection() does, but uses the "normal"
	libq global variable interfaces instead of arguments. */

int
_libpq_EstablishComm()
{
  extern void PQufs_init();
  libpq_connection_t cid, other_cid;
  u_short portno;
  extern char PQdatabase[], *PQoption, *PQhost, *PQtty, *PQport;
  libpq_per_thread_state_t *s;

  portno = atoi(PQport);
  s = _libpq_setup_thread();
  cid = s->cid;
  other_cid = _libpq_find_connection(PQhost,portno,PQdatabase);
  if ((cid == INVALID_CONN) || (cid != other_cid)) {
    cid = pq_connect(PQdatabase,
		     fe_getauthname(),
		     PQoption,
		     PQhost,
		     PQtty,
		     (char *) NULL,
		     (short) atoi(PQport));
    if (cid == INVALID_CONN) {
      (void) sprintf(s->PQerrormsg, 
		     "FATAL: Failed to connect to postmaster (host=%s, port=%s)\n",
		     PQhost, PQport);
      pqdebug("%s", s->PQerrormsg);
      (void) strcat(s->PQerrormsg, "\tIs the postmaster running?\n");
      fputs(s->PQerrormsg, stderr);
      cid = INVALID_CONN;
    } else {
      pq_flush();
      PQufs_init();
    }
  }
  if (cid != INVALID_CONN)
    _libpq_update_use_time(cid);
  return cid;
}


/* ----------------------------------------
 * pq_connect  -- initiate a communication link between client and
 *	POSTGRES backend via postmaster.
 *
 * RETURNS: STATUS_ERROR, if arguments are wrong or local communication
 *	status is screwed up (can't create socket, etc).  STATUS_OK
 *	otherwise.
 *
 * SIDE_EFFECTS: initiates connection.  
 *               SIGURG handler is set (async notification)
 *
 * NOTE: we don't wait for any error messages from the backend/postmaster.
 *	That means that if the fork fails or the startup message is corrupted,
 *	we won't find out until the first Send/Receive message cal.
 *
 *    XXX This is a thread-aware version of the routine by the same
 *	  name in pqcomm.c, which is ifdef'ed out if USE_THREADS is
 *	  turned on.
 * ----------------------------------------
 */

libpq_connection_t
pq_connect(dbname,user,args,hostName,debugTty,execFile,portName)
     char	*dbname;
     char	*user;
     char	*args;
     char	*hostName;
     char	*debugTty;
     char	*execFile;
     short	portName;
{
  /*
   * This data structure is used for the seq-packet protocol.  It
   * describes the frontend-backend connection.
   */
  Connection *MyConn = NULL;
  Port *SendPort = NULL;
  StartupPacket startup;
  int status;
  MsgType msgtype;
  libpq_connection_t cid;
  libpq_per_thread_state_t *s;
  extern char PQdatabase[], *PQoption, *PQhost, *PQtty, *PQport;
  int have_lock = 0;

  s = _libpq_setup_thread();
  status = STATUS_OK;
  GET_LOCK(connection_table_lock);
  have_lock = 1;
  cid = _libpq_find_connection(hostName, portName, dbname);
  if (cid != INVALID_CONN) {
    REL_LOCK(connection_table_lock);
    have_lock = 0;
    _libpq_assign_connection(cid);
    goto done;
  }

  /*
   * Initialize the startup packet.  Packet fields defined in comm.h
   */
  strncpy(startup.database,dbname,sizeof(startup.database));
  strncpy(startup.user,user,sizeof(startup.user));
  strncpy(startup.options,args,sizeof(startup.options));
  strncpy(startup.tty,debugTty,sizeof(startup.tty));
  if (execFile != NULL) {
    strncpy(startup.execFile, execFile, sizeof(startup.execFile));
  } else {
    strncpy(startup.execFile, "", sizeof(startup.execFile));
  }

  /* If no port  was suggested grab the default or PGPORT value */
  if (!portName)
    portName = pq_getport();
  /*
   * initialize connection structure.  This is really needed
   * only for the sequenced packet protocol, but these initializations
   * are important to the packet.c library.
   */
  MyConn = (Connection *) malloc(sizeof(Connection));
  bzero((char*)MyConn, sizeof(Connection));
  MyConn->id = INVALID_ID;
  MyConn->seqno = INITIAL_SEQNO;

  /*
   * Open a connection to postmaster/backend.
   */
  SendPort = (Port *) malloc(sizeof(Port));
  bzero((char *) SendPort, sizeof(Port));
  status = StreamOpen(hostName, portName, SendPort);
  if (status != STATUS_OK)
    goto done;

  /* initialize */
  msgtype = fe_getauthsvc();
  status = PacketSend(SendPort, MyConn, &startup, msgtype,
		      sizeof(startup), BLOCKING);

  /* authenticate as required */
  if (fe_sendauth(msgtype, SendPort, hostName) != STATUS_OK) {
    (void) sprintf(s->PQerrormsg,
		   "FATAL: pq_connect: authentication failed with %s\n",
		   hostName);
    fputs(s->PQerrormsg, stderr);
    pqdebug("%s", s->PQerrormsg);
    status = STATUS_ERROR;
    goto done;
  }

  /* set up the socket file descriptors */
  cid = _libpq_new_connection();
  if (cid == INVALID_CONN) {
    status = STATUS_ERROR;
    goto done;
  }
  bcopy((char *)&SendPort->raddr, (char *)&connections[cid]._addr,
	sizeof(struct sockaddr_in));
  connections[cid]._dbname = _libpq_get_dbname(dbname);
  connections[cid]._dont_reap = PQpersistentConns;
  connections[cid].PQhost = dup_str(hostName);
  { char portStr[64];

    sprintf(portStr, "%d", (int)portName);
    connections[cid].PQport = dup_str(portStr);
  }
  connections[cid].PQtty = dup_str(PQtty);
  connections[cid].PQoption = dup_str(PQoption);
  connections[cid].Pfout = fdopen(SendPort->sock, "w");
  connections[cid].Pfin = fdopen(dup(SendPort->sock), "r");
  strncpy(connections[cid].PQdatabase, dbname, 16);
  if (!connections[cid].Pfout || !connections[cid].Pfin) {
    (void) sprintf(s->PQerrormsg,
		   "FATAL: pq_connect: fdopen() failed: errno=%d\n",
		   errno);
    fputs(s->PQerrormsg, stderr);
    pqdebug("%s", s->PQerrormsg);
    return(STATUS_ERROR);
  }

  connections[cid].PQAsyncNotifyWaiting = 0;
  connections[cid].PQportset = 1;
  REL_LOCK(connection_table_lock);
  have_lock = 0;
  _libpq_assign_connection(cid);

  PQnotifies_init();

  /*    pq_regoob(pq_async_notify);*/

 done:
  if (MyConn)
    free((void *)MyConn);
  if (SendPort)
    free((void *)SendPort);
  if (have_lock)
    REL_LOCK(connection_table_lock);
  return (status == STATUS_OK) ? cid : INVALID_CONN;
}

/* pq_close -- let go of a connection

	thread-aware version of routine in pqcomm.c. we don't
	actually shut down the connection, just release it for
	use by other threads. */

void
pq_close()
{
  libpq_connection_t cid;

  cid = _libpq_current_connection();
  if (cid != INVALID_CONN) {
    _libpq_release_connection(cid);
  }
}
#endif /* USE_THREADS */

/* All of the following serve to hide the per-thread state from the
   rest of the world, and collude with the macros in pqthreads-internal.h
   to fool the rest of the global-infected code in this library into
   thinking it owns the universe. */

#ifdef __STDC__
#define DefinePerThreadStateAccessor(T,M)			\
T _libpq_get_##M()						\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  return s? s->M : (T)0;					\
}
#define DefinePerThreadStateSetter(T,M)				\
void _libpq_set_##M(T v)					\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  if (s)							\
    s->M = v;							\
}
#define DefinePerConnectionStateAccessor(T,M)			\
T _libpq_get_##M()						\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  return s? ((s->cid!=INVALID_CONN)?connections[s->cid].M :	\
		(T)0) : (T)0;					\
}
#define DefinePerConnectionStateSetter(T,M)			\
void _libpq_set_##M(T v)					\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  if (s && (s->cid != INVALID_CONN))				\
    connections[s->cid].M=v;					\
}
#else
#define DefinePerThreadStateAccessor(T,M)			\
T _libpq_get_/**/M()						\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  return s? s->M : (T)0;					\
}
#define DefinePerThreadStateSetter(T,M)				\
void _libpq_set_/**/M(v)					\
     T v;							\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  if (s)							\
    s->M = v;							\
}
#define DefinePerConnectionStateAccessor(T,M)			\
T _libpq_get_/**/M()						\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  return s? ((s->cid!=INVALID_CONN)?connections[s->cid].M :	\
		(T)0) : (T)0;					\
}
#define DefinePerConnectionStateSetter(T,M)			\
void _libpq_set_/**/M(v)					\
     T v;							\
{ libpq_per_thread_state_t *s;					\
  s = GET_PER_THREAD_STATE(libpq_state_key);			\
  if (s && (s->cid != INVALID_CONN))				\
    connections[s->cid].M=v;					\
}
#endif /* __STDC__ */

DefinePerConnectionStateAccessor(FILE*,Pfout)
DefinePerConnectionStateAccessor(FILE*,Pfin)
DefinePerConnectionStateAccessor(int,PQAsyncNotifyWaiting)
DefinePerConnectionStateSetter(int,PQAsyncNotifyWaiting)
DefinePerConnectionStateAccessor(char*,PQhost)
DefinePerConnectionStateAccessor(char*,PQport)
DefinePerConnectionStateAccessor(char*,PQtty)
DefinePerConnectionStateAccessor(char*,PQoption)
DefinePerConnectionStateAccessor(char*,PQdatabase)
DefinePerConnectionStateAccessor(int,PQportset)
DefinePerConnectionStateAccessor(int,PQxactid)
DefinePerConnectionStateSetter(int,PQxactid)
DefinePerThreadStateAccessor(char*,retbuf)
DefinePerThreadStateAccessor(char*,cwdir)
DefinePerThreadStateAccessor(char*,thenewdir)
DefinePerThreadStateAccessor(int,p_errno)
DefinePerThreadStateSetter(int,p_errno)
DefinePerConnectionStateAccessor(int,PQtracep)
DefinePerConnectionStateSetter(int,PQtracep)
DefinePerThreadStateAccessor(char*,PQerrormsg)
DefinePerThreadStateAccessor(FILE*,debug_port)
DefinePerThreadStateSetter(FILE*,debug_port)
